/********** Fichiers d'inclusion **************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>

#include <arpa/nameser.h>
#include <resolv.h>
#include <errno.h>

#include "lib_SMTP.h"
#include "../Communication/libcom.h"


/********** Fonctions d'envoi des mail au serveur de destination **********/

/* --- Fonction qui permet de lire message envoye par le serveur de destination ---*/
static void lit_repServeur(int sock, char *buffer)
{
	char *p;
    int n = 0;

    if((n = recv(sock, buffer, BUF_SIZE - 1, 0)) < 0)
    {
        perror("recv()");
        exit(errno);
    }

    if(n == 0)
    {
        printf("Connection lost\n");
      	exit(EXIT_FAILURE);
    }

    buffer[n] = 0;

    p = strstr(buffer, "\r\n");

    if(p != NULL)
		*p = 0;
}


/* --- Fonction d'envoi de message au serveur SMTP de destination --- */
static void envoie_serveur(int sock, char *buffer)
{
    if(send(sock, buffer, strlen(buffer), 0) < 0)
    {
        perror("send()");
        exit(errno);
    }
}


/* --- Fonction qui se connecte et envoie le mail au serveur SMTP de destination --- */
int envoie_mail(Client_t *client,struct struct_mx *mx)
{
	/* Résoudre l'IP du serveur de destination avec la commande "host" */
	//char command[MAX_LIGNE];
	//strcpy(command,"host ");
	//strcat(command,mx[0].servername);
	/* Stocker l'IP du serveur de destination dans un fichier "mxrecord.txt" */
	//strcat(command, " | cut -f4 -d ' ' >mxrecord.txt");
	//system(command);
	
	int sock = connexionServeur(mx[0].servername,25);
	if(sock<0)
	{
		fprintf(stderr,"Erreur de connexion au serveur\n");
		exit(EXIT_FAILURE);
	}
	
	char command[BUF_SIZE];
	char buffer[BUF_SIZE];

	/* Lire la rep du serveur SMTP de destination */
	lit_repServeur(sock, buffer);
	puts(buffer);
	
	/* Envoi de HELO */
	sprintf(command, "HELO %s\r\n", client->helo);
	//puts(command);
	envoie_serveur(sock, command);

	/* Lire reponse */
	lit_repServeur(sock, buffer);
    puts(buffer);
		
	/* Envoi de MAIL FROM */
	sprintf(command, "MAIL FROM:<%s>\r\n", client->mail);
	//puts(command);
	envoie_serveur(sock, command);

	/* Lire reponse */
    lit_repServeur(sock, buffer);
    puts(buffer);

	/* Envoi de RCPT TO */
    sprintf(command, "RCPT TO:<%s>\r\n", client->rcpt);
    //puts(command);
    envoie_serveur(sock, command);

	/* Lire reponse */
    lit_repServeur(sock, buffer);
    puts(buffer);
	
	/* Envoi DATA */
	sprintf(command,"DATA\r\n");
	//puts(command);
	envoie_serveur(sock, command);
	
	/* Lire reponse */
    lit_repServeur(sock, buffer);
    puts(buffer);
	
	/* Envoi du corps du message */
	sprintf(command,"%s\n", (char *)client->body);
	//puts(client->body);
	envoie_serveur(sock, command);
	
	/* Envoi de la marque de fin du message */
	envoie_serveur(sock, ".\n");

	/* Lire reponse */
    lit_repServeur(sock, buffer);
    puts(buffer);
	
	/* Terminer le dialogue avec le serveur */
	envoie_serveur(sock, "QUIT\r\n");

	 /* Lire reponse */
        lit_repServeur(sock, buffer);
        puts(buffer);

	//fclose(dialogue);
	shutdown(sock,SHUT_RDWR);
	
	return 0;
}







/********** Fonctions de gestion des MX **********/

/* --- Fonction qui compare la priorité des serveurs de destination --- */
static int compare (const void *a, const void *b)
{  
    if (((mx_t *)a)->priority < ((mx_t *)b)->priority) 
    {
        return -1;
	}
	else if (((mx_t *)a)->priority > ((mx_t *)b)->priority)
	{
        return 1;
    } 
	else
	{
		return 0;
    }
}

/* --- Fonction qui affiche les noms des serveurs de destination --- */
static void aff (struct struct_mx *mx, size_t n)
{
   size_t i; 
   for (i = 0; i <= n; i++)
   {
      printf ("---> %d     %s \n\n", mx[i].priority,mx[i].servername);
   }
}

/* --- Fonction qui résoud les domaines MX en noms des serveurs --- */
int resoud_mx (Client_t* client, struct struct_mx *mx)
{
	unsigned char Buffer[8000]   = {0};
	const unsigned char *pResult = NULL;
	//struct struct_mx mx[16];
	struct __res_state ResState  = {0};
	ns_msg nsMsg = {0};
	ns_rr  rr;

	int  type     = 0;
	int  ret      = 0;
	int  size     = 0;
	int  count    = 0;
	int  i        = 0;
	int  res_init = 0;

	ret = res_ninit (&ResState);

	if (ret)
	{
		printf ("error initializing resolver\n");
		goto close;
	}
	res_init = 1;

	memset(Buffer, 0, sizeof (Buffer));

	size = res_nquery (&ResState, client->domaine, C_IN, T_MX, Buffer, sizeof(Buffer)-1);

	if (0 == size)
		goto close;

	ret = ns_initparse (Buffer, size, &nsMsg);

	if (ret)
		goto close;

	count = ns_msg_count (nsMsg, ns_s_an);

	for (i=0; i< count;i++)
	{
		ret = ns_parserr (&nsMsg, ns_s_an, i , &rr);

		if (ret)
			goto close;

		type = ns_rr_type (rr);
		
		if (ns_t_mx == type)
		{
			pResult = ns_rr_rdata (rr);
			NS_GET16(mx[i].priority,pResult);
			dn_expand(ns_msg_base(nsMsg),ns_msg_end(nsMsg),pResult,mx[i].servername,MAX_SERVER_NAME);
			//printf("%d %s\n",mx[i].priority,mx[i].servername);
	    }
	}
	
    qsort (mx, count, sizeof (mx_t), compare);
    aff (mx, count);
	
	close:
		if (res_init)
			res_nclose (&ResState);
	
	return 0;
}


////////////////////
void trouve_mx(Client_t* client,FILE* dialogue){
    int i=0;
    int flag_fin=0;
    int flag_arobase=0;

    while (flag_fin==0)
    {
        if (client->rcpt[i]!='@' && flag_arobase ==0)
        {
            i++;
        }
        else
        {
            flag_arobase =1;
            i++;
            for (int j=0; j<100; j++)
            {
                if (client->rcpt[i] != '\0')
                {
                    client->domaine[j]=client->rcpt[i];
                    //fprintf(dialogue,"domaine intermediaire:%s\r\n",client->domaine);
                    i++;
                }
                else
                {
                    //fprintf(dialogue,"domaine final:%s\r\n",client->domaine);
                    flag_fin=1;
                    break;
                }
            }
        }
    }
}







/******************** Fonctions de gestion SMTP ********************/

/* --- Fonction qui récupère et stocke le nom d'utilisateur de l'expéditeur --- */
void gestion_HELO(char *ligne,char *code,Client_t* client,FILE* dialogue)
{
	sscanf(ligne,"%4s %s",code,client->helo);
	fprintf(dialogue,"250 TristanJerome ZABETH11 (ID:%s) \r\n",client->helo);
}

/* --- Fonction qui récupère et stocke l'email de l'expéditeur --- */
void gestion_MAIL(char *ligne,char *code,Client_t* client,FILE* dialogue)
{
	sscanf(ligne,"%4s %s",code,client->mail);
	client->mail[strlen(client->mail)+1]='\0';
	fprintf(dialogue,"250 MAIL FROM: %s\r\n",client->mail);
}

/* --- Fonction qui récupère et stocke l'email du destinataire --- */
void gestion_RCPT(char *ligne,char *code,Client_t* client,FILE* dialogue)
{
	sscanf(ligne,"%4s %s",code,client->rcpt);
	client->rcpt[strlen(client->rcpt)+1]='\0';
	fprintf(dialogue,"250 RCPT TO: %s\r\n",client->rcpt);
}

/* --- Fonction qui récupère le champ DATA renseigné par le client  --- */
void gestion_DATA(char *ligne, char *code, Client_t* client, FILE* dialogue)
{
   	client->taille_max=MAX_ALLOC;
	client->taille_body=0;
	client->body = malloc(client->taille_max);
	client->body[0]='\0';
	
	while(fgets(ligne,MAX_LIGNE,dialogue)!=NULL)
	{
		char verif[MAX_LIGNE];
		sscanf(ligne, "%s", verif);
		if (strcmp(verif,".")==0)
		{
			fprintf(dialogue,"250 DATA\r\n");
			printf("\n\r%s \n\r",client->body);
			break;
		}
		else if ((client->taille_max-client->taille_body)<=strlen(ligne))
		{		
			client->taille_max += MAX_ALLOC;
			client->body=realloc(client->body,client->taille_max);
		}
		strcat(client->body,ligne);
		client->taille_body += strlen(ligne);
	}
}

//////////
void gestion_QUIT(FILE* dialogue)
{
    fprintf(dialogue,"221 Bye\r\n");
}

/* --- Fonction qui gère le protocole SMTP pour les clients  --- */
int gestionSmtp(void * arg)
{
	/* Socket */
	int s = *(int *) arg;
	
	/* Obtenir une structure de fichier */
	Client_t client;
	
	FILE *dialogue=fdopen(s,"a+");
	if(dialogue==NULL)
	{ 
		perror("gestionClient.fdopen");
		exit(EXIT_FAILURE);
	}
	else {fprintf(dialogue,"220 TristanJerome SMTP\n");}
	
	int flag_helo = 0;
	int flag_mail = 0;
	int flag_rcpt = 0;
	int flag_data = 0;
	//int flag_fin = 0;

	char code[MAX_CODE];
	char ligne[MAX_LIGNE];

	while(fgets(ligne,MAX_LIGNE,dialogue)!=NULL)
	{
		sscanf(ligne,"%4s",code);
		if (strcmp(code,"HELO")==0)
		{	
			if (flag_helo < 1)
			{
				gestion_HELO(ligne,code,&client,dialogue);
				flag_helo = 1;
				//flag_fin ++;
			}
		}

		else if(strcmp(code,"MAIL")==0)
		{
			if(flag_helo == 1 && flag_mail < 1)
			{
				gestion_MAIL(ligne,code,&client,dialogue);
				flag_mail = 1;
				//flag_fin ++;
			}
			else {fprintf(dialogue,"503 Les personnes polies s'annoncent\r\n");}
		}
	
		else if(strcmp(code,"RCPT")==0)
		{	
			if (flag_helo==1 && flag_rcpt < 1)
			{
				gestion_RCPT(ligne,code,&client,dialogue);
				flag_rcpt =1;
				//flag_fin ++;
			}
			else {fprintf(dialogue,"503 Les personnes polies s'annoncent\r\n");}
		}

		else if(strcmp(code,"DATA")==0)
		{
			if (flag_helo==1 && flag_data < 1)
			{
				fprintf(dialogue,"354 \n\r");
				gestion_DATA(ligne,code,&client,dialogue);
				flag_data = 1;
				//flag_fin ++;
			}
			else {fprintf(dialogue,"503 Les personnes polies s'annoncent\r\n");}
		}

		else if (strcmp(code,"QUIT")==0) 
		{
			gestion_QUIT(dialogue);
			break;
		}
		
		else fprintf(dialogue, "502 commande non reconnue par le système, contacter le programmeur\r\n");
	}
	
	/* Trouve les serveurs associés au nom de domaine du mx */
	trouve_mx(&client,dialogue);
	
	/* Résoud l'IP des serveurs */
	struct struct_mx mx[16];
	resoud_mx(&client,mx);
	
	/* Envoie mail au serveur de destination*/
	envoie_mail(&client,mx);
	
	/* Termine la connexion */
	fclose(dialogue);
	
	return 0;
}